@@ -6,7 +6,6 @@
 #include <asm/cputable.h>
 
 #include "asm/types.h"
-#include "asm/cpu.h"
 
 #include "cr_options.h"
 #include "image.h"
@@ -17,7 +16,7 @@
 #include "protobuf.h"
 #include "images/cpuinfo.pb-c.h"
 
-static uint64_t hwcap[2];
+static compel_cpuinfo_t rt_cpuinfo;
 
 #ifdef __LITTLE_ENDIAN__
 #define CURRENT_ENDIANNESS	CPUINFO_PPC64_ENTRY__ENDIANNESS__LITTLEENDIAN
@@ -27,14 +26,7 @@ static uint64_t hwcap[2];
 
 int cpu_init(void)
 {
-	hwcap[0] = getauxval(AT_HWCAP);
-	hwcap[1] = getauxval(AT_HWCAP2);
-
-	if (!hwcap[0] || !hwcap[1]) {
-		pr_err("Can't read the hardware capabilities");
-		return -1;
-	}
-	return 0;
+	return compel_cpuid(&rt_cpuinfo);
 }
 
 int cpu_dump_cpuinfo(void)
@@ -46,20 +38,20 @@ int cpu_dump_cpuinfo(void)
 	int ret = -1;
 
 	img = open_image(CR_FD_CPUINFO, O_DUMP);
-        if (!img)
-                return -1;
+	if (!img)
+		return -1;
 
 	cpu_info.ppc64_entry = &cpu_ppc64_info_ptr;
 	cpu_info.n_ppc64_entry = 1;
 
 	cpu_ppc64_info.endian = CURRENT_ENDIANNESS;
 	cpu_ppc64_info.n_hwcap = 2;
-	cpu_ppc64_info.hwcap = hwcap;
+	cpu_ppc64_info.hwcap = rt_cpuinfo.hwcap;
 
 	ret = pb_write_one(img, &cpu_info, PB_CPUINFO);
 
 	close_image(img);
-        return ret;
+	return ret;
 }
 
 int cpu_validate_cpuinfo(void)
@@ -68,21 +60,21 @@ int cpu_validate_cpuinfo(void)
 	CpuinfoPpc64Entry *cpu_ppc64_entry;
 	struct cr_img *img;
 	int ret = -1;
-        img = open_image(CR_FD_CPUINFO, O_RSTR);
-        if (!img)
-                return -1;
+	img = open_image(CR_FD_CPUINFO, O_RSTR);
+	if (!img)
+		return -1;
 
-        if (pb_read_one(img, &cpu_info, PB_CPUINFO) < 0)
-                goto error;
+	if (pb_read_one(img, &cpu_info, PB_CPUINFO) < 0)
+		goto error;
 
 	if (cpu_info->n_ppc64_entry != 1) {
-		pr_err("No PPC64 related entry in image");
+		pr_err("No PPC64 related entry in image\n");
 		goto error;
 	}
 	cpu_ppc64_entry = cpu_info->ppc64_entry[0];
 
 	if (cpu_ppc64_entry->endian != CURRENT_ENDIANNESS) {
-		pr_err("Bad endianness");
+		pr_err("Bad endianness\n");
 		goto error;
 	}
 
@@ -92,7 +84,8 @@ int cpu_validate_cpuinfo(void)
 	}
 
 #define CHECK_FEATURE(s,f) do {						\
-		if ((cpu_ppc64_entry->hwcap[s] & f) && !(hwcap[s] & f)) { \
+		if ((cpu_ppc64_entry->hwcap[s] & f) &&			\
+		    !(rt_cpuinfo.hwcap[s] & f)) {			\
 			pr_err("CPU Feature %s required by image "	\
 			       "is not supported on host.\n", #f);	\
 			goto error;					\
@@ -128,22 +121,22 @@ error:
 
 int cpuinfo_dump(void)
 {
-        if (cpu_init())
-                return -1;
+	if (cpu_init())
+		return -1;
 
-        if (cpu_dump_cpuinfo())
-                return -1;
+	if (cpu_dump_cpuinfo())
+		return -1;
 
-        return 0;
+	return 0;
 }
 
 int cpuinfo_check(void)
 {
-        if (cpu_init())
-                return -1;
+	if (cpu_init())
+		return -1;
 
-        if (cpu_validate_cpuinfo())
-                return 1;
+	if (cpu_validate_cpuinfo())
+		return 1;
 
 	return 0;
 }
