@@ -12,19 +12,16 @@
  * This is a minimal amount of symbols
  * we should support at the moment.
  */
-#define VDSO_SYMBOL_MAX	4
+#define VDSO_SYMBOL_MAX	7
 
 #define ARCH_VDSO_SYMBOLS			\
 	"__vdso_clock_gettime",			\
 	"__vdso_getcpu",			\
 	"__vdso_gettimeofday",			\
-	"__vdso_time"
+	"__vdso_time",				\
+	"__kernel_vsyscall",			\
+	"__kernel_sigreturn",			\
+	"__kernel_rt_sigreturn"
 
 
-struct vdso_symtable;
-extern int vdso_redirect_calls(unsigned long base_to,
-			       unsigned long base_from,
-			       struct vdso_symtable *to,
-			       struct vdso_symtable *from);
-
 #endif /* __CR_ASM_VDSO_H__ */
