@@ -13,8 +13,7 @@
 #include <sys/epoll.h>
 
 #include "crtools.h"
-#include "compiler.h"
-#include "asm/types.h"
+#include "common/compiler.h"
 #include "imgset.h"
 #include "rst_info.h"
 #include "eventpoll.h"
@@ -22,6 +21,7 @@
 #include "image.h"
 #include "util.h"
 #include "log.h"
+#include "pstree.h"
 
 #include "protobuf.h"
 #include "images/eventpoll.pb-c.h"
@@ -34,13 +34,6 @@ struct eventpoll_file_info {
 	struct file_desc		d;
 };
 
-struct eventpoll_tfd_file_info {
-	EventpollTfdEntry		*tdefe;
-	struct list_head		list;
-};
-
-static LIST_HEAD(eventpoll_tfds);
-
 /* Checks if file descriptor @lfd is eventfd */
 int is_eventpoll_link(char *link)
 {
@@ -58,53 +51,32 @@ static void pr_info_eventpoll(char *action, EventpollFileEntry *e)
 	pr_info("%seventpoll: id %#08x flags %#04x\n", action, e->id, e->flags);
 }
 
-struct eventpoll_list {
-	struct list_head list;
-	int n;
-};
-
-static int dump_eventpoll_entry(union fdinfo_entries *e, void *arg)
-{
-	struct eventpoll_list *ep_list = (struct eventpoll_list *) arg;
-	EventpollTfdEntry *efd = &e->epl.e;
-
-	pr_info_eventpoll_tfd("Dumping: ", efd);
-
-	list_add_tail(&e->epl.node, &ep_list->list);
-	ep_list->n++;
-
-	return 0;
-}
-
 static int dump_one_eventpoll(int lfd, u32 id, const struct fd_parms *p)
 {
+	FileEntry fe = FILE_ENTRY__INIT;
 	EventpollFileEntry e = EVENTPOLL_FILE_ENTRY__INIT;
-	struct eventpoll_list ep_list = {LIST_HEAD_INIT(ep_list.list), 0};
-	union fdinfo_entries *te, *tmp;
 	int i, ret = -1;
 
 	e.id = id;
 	e.flags = p->flags;
 	e.fown = (FownEntry *)&p->fown;
 
-	if (parse_fdinfo(lfd, FD_TYPES__EVENTPOLL, dump_eventpoll_entry, &ep_list))
+	if (parse_fdinfo(lfd, FD_TYPES__EVENTPOLL, &e))
 		goto out;
 
-	e.tfd = xmalloc(sizeof(struct EventpollTfdEntry *) * ep_list.n);
-	if (!e.tfd)
-		goto out;
-
-	i = 0;
-	list_for_each_entry(te, &ep_list.list, epl.node)
-		e.tfd[i++] = &te->epl.e;
-	e.n_tfd = ep_list.n;
+	fe.type = FD_TYPES__EVENTPOLL;
+	fe.id = e.id;
+	fe.epfd = &e;
 
 	pr_info_eventpoll("Dumping ", &e);
-	ret = pb_write_one(img_from_set(glob_imgset, CR_FD_EVENTPOLL_FILE),
-		     &e, PB_EVENTPOLL_FILE);
+	ret = pb_write_one(img_from_set(glob_imgset, CR_FD_FILES), &fe, PB_FILE);
 out:
-	list_for_each_entry_safe(te, tmp, &ep_list.list, epl.node)
-		free_event_poll_entry(te);
+	for (i = 0; i < e.n_tfd; i++) {
+		if (!ret)
+			pr_info_eventpoll_tfd("Dumping: ", e.tfd[i]);
+		eventpoll_tfd_entry__free_unpacked(e.tfd[i], NULL);
+	}
+	xfree(e.tfd);
 
 	return ret;
 }
@@ -114,13 +86,19 @@ const struct fdtype_ops eventpoll_dump_ops = {
 	.dump		= dump_one_eventpoll,
 };
 
-static int eventpoll_open(struct file_desc *d)
+static int eventpoll_post_open(struct file_desc *d, int fd);
+
+static int eventpoll_open(struct file_desc *d, int *new_fd)
 {
+	struct fdinfo_list_entry *fle = file_master(d);
 	struct eventpoll_file_info *info;
 	int tmp;
 
 	info = container_of(d, struct eventpoll_file_info, d);
 
+	if (fle->stage >= FLE_OPEN)
+		return eventpoll_post_open(d, fle->fe->fd);
+
 	pr_info_eventpoll("Restore ", info->efe);
 
 	tmp = epoll_create(1);
@@ -136,11 +114,34 @@ static int eventpoll_open(struct file_desc *d)
 		goto err_close;
 	}
 
-	return tmp;
+	*new_fd = tmp;
+	return 1;
 err_close:
 	close(tmp);
 	return -1;
 }
+
+static int epoll_not_ready_tfd(EventpollTfdEntry *tdefe)
+{
+	struct fdinfo_list_entry *fle;
+
+	list_for_each_entry(fle, &rsti(current)->fds, ps_list) {
+		if (tdefe->tfd != fle->fe->fd)
+			continue;
+
+		if (fle->desc->ops->type == FD_TYPES__EVENTPOLL)
+			return (fle->stage < FLE_OPEN);
+		else
+			return (fle->stage != FLE_RESTORED);
+	}
+
+	/*
+	 * If tgt fle is not on the fds list, it's already
+	 * restored (see open_fdinfos), so we're ready.
+	 */
+	return 0;
+}
+
 static int eventpoll_retore_tfd(int fd, int id, EventpollTfdEntry *tdefe)
 {
 	struct epoll_event event;
@@ -159,52 +160,55 @@ static int eventpoll_retore_tfd(int fd, int id, EventpollTfdEntry *tdefe)
 
 static int eventpoll_post_open(struct file_desc *d, int fd)
 {
-	struct eventpoll_tfd_file_info *td_info;
 	struct eventpoll_file_info *info;
 	int i;
 
 	info = container_of(d, struct eventpoll_file_info, d);
 
+	for (i = 0; i < info->efe->n_tfd; i++) {
+		if (epoll_not_ready_tfd(info->efe->tfd[i]))
+			return 1;
+	}
 	for (i = 0; i < info->efe->n_tfd; i++) {
 		if (eventpoll_retore_tfd(fd, info->efe->id, info->efe->tfd[i]))
 			return -1;
 	}
 
-	list_for_each_entry(td_info, &eventpoll_tfds, list) {
-		if (td_info->tdefe->id != info->efe->id)
-			continue;
-
-		if (eventpoll_retore_tfd(fd, info->efe->id, td_info->tdefe))
-			return -1;
-
-	}
-
 	return 0;
 }
 
-static void eventpoll_collect_fd(struct file_desc *d,
-		struct fdinfo_list_entry *fle, struct rst_info *ri)
-{
-	list_add_tail(&fle->ps_list, &ri->eventpoll);
-}
-
 static struct file_desc_ops desc_ops = {
 	.type = FD_TYPES__EVENTPOLL,
 	.open = eventpoll_open,
-	.post_open = eventpoll_post_open,
-	.collect_fd = eventpoll_collect_fd,
 };
 
 static int collect_one_epoll_tfd(void *o, ProtobufCMessage *msg, struct cr_img *i)
 {
-	struct eventpoll_tfd_file_info *info = o;
+	EventpollTfdEntry *tfde;
+	struct file_desc *d;
+	struct eventpoll_file_info *ef;
+	EventpollFileEntry *efe;
+	int n_tfd;
 
 	if (!deprecated_ok("Epoll TFD image"))
 		return -1;
 
-	info->tdefe = pb_msg(msg, EventpollTfdEntry);
-	list_add(&info->list, &eventpoll_tfds);
-	pr_info_eventpoll_tfd("Collected ", info->tdefe);
+	tfde = pb_msg(msg, EventpollTfdEntry);
+	d = find_file_desc_raw(FD_TYPES__EVENTPOLL, tfde->id);
+	if (!d) {
+		pr_err("No epoll FD for %u\n", tfde->id);
+		return -1;
+	}
+
+	ef = container_of(d, struct eventpoll_file_info, d);
+	efe = ef->efe;
+
+	n_tfd = efe->n_tfd + 1;
+	if (xrealloc_safe(&efe->tfd, n_tfd * sizeof(EventpollTfdEntry *)))
+		return -1;
+
+	efe->tfd[efe->n_tfd] = tfde;
+	efe->n_tfd = n_tfd;
 
 	return 0;
 }
@@ -212,8 +216,8 @@ static int collect_one_epoll_tfd(void *o, ProtobufCMessage *msg, struct cr_img *
 struct collect_image_info epoll_tfd_cinfo = {
 	.fd_type = CR_FD_EVENTPOLL_TFD,
 	.pb_type = PB_EVENTPOLL_TFD,
-	.priv_size = sizeof(struct eventpoll_tfd_file_info),
 	.collect = collect_one_epoll_tfd,
+	.flags = COLLECT_NOFREE,
 };
 
 static int collect_one_epoll(void *o, ProtobufCMessage *msg, struct cr_img *i)
