@@ -13,7 +13,7 @@ static int dump_one_ext_file(int lfd, u32 id, const struct fd_parms *p)
 {
 	int ret;
 	struct cr_img *rimg;
-
+	FileEntry fe = FILE_ENTRY__INIT;
 	ExtFileEntry xfe = EXT_FILE_ENTRY__INIT;
 
 	ret = run_plugins(DUMP_EXT_FILE, lfd, id);
@@ -23,8 +23,12 @@ static int dump_one_ext_file(int lfd, u32 id, const struct fd_parms *p)
 	xfe.id		= id;
 	xfe.fown	= (FownEntry *)&p->fown;
 
-	rimg = img_from_set(glob_imgset, CR_FD_EXT_FILES);
-	return pb_write_one(rimg, &xfe, PB_EXT_FILE);
+	fe.type = FD_TYPES__EXT;
+	fe.id = xfe.id;
+	fe.ext = &xfe;
+
+	rimg = img_from_set(glob_imgset, CR_FD_FILES);
+	return pb_write_one(rimg, &fe, PB_FILE);
 }
 
 const struct fdtype_ops ext_dump_ops = {
@@ -37,7 +41,7 @@ struct ext_file_info {
 	ExtFileEntry		*xfe;
 };
 
-static int open_fd(struct file_desc *d)
+static int open_fd(struct file_desc *d, int *new_fd)
 {
 	struct ext_file_info *xfi;
 	int fd;
@@ -53,7 +57,8 @@ static int open_fd(struct file_desc *d)
 	if (restore_fown(fd, xfi->xfe->fown))
 		return -1;
 
-	return fd;
+	*new_fd = fd;
+	return 0;
 }
 
 static struct file_desc_ops ext_desc_ops = {
@@ -79,11 +84,11 @@ struct collect_image_info ext_file_cinfo = {
 };
 
 int dump_unsupp_fd(struct fd_parms *p, int lfd,
-			  struct cr_img *img, char *more, char *info)
+		char *more, char *info, FdinfoEntry *e)
 {
 	int ret;
 
-	ret = do_dump_gen_file(p, lfd, &ext_dump_ops, img);
+	ret = do_dump_gen_file(p, lfd, &ext_dump_ops, e);
 	if (ret == 0)
 		return 0;
 	if (ret == -ENOTSUP)
