@@ -1,8 +1,7 @@
 #ifndef __CR_CGROUP_H__
 #define __CR_CGROUP_H__
 
-#include "asm/int.h"
-
+#include "int.h"
 #include "images/core.pb-c.h"
 
 struct pstree_item;
