@@ -49,4 +49,8 @@
 #define BINFMTFS_MAGIC		0x42494e4d
 #endif
 
+#ifndef AUTOFS_SUPER_MAGIC
+#define AUTOFS_SUPER_MAGIC	0x0187
+#endif
+
 #endif /* __CR_FS_MAGIC_H__ */
