@@ -1,7 +1,7 @@
 #ifndef __CR_IMAGE_DESC_H__
 #define __CR_IMAGE_DESC_H__
 
-#include "asm/int.h"
+#include "int.h"
 
 enum {
 	CR_FD_INVENTORY,
@@ -14,7 +14,6 @@ enum {
 	CR_FD_CORE,
 	CR_FD_IDS,
 	CR_FD_MM,
-	CR_FD_SIGACT,
 	CR_FD_CREDS,
 	CR_FD_FS,
 	_CR_FD_TASK_TO,
@@ -55,30 +54,14 @@ enum {
 	CR_FD_FDINFO,
 
 	_CR_FD_GLOB_FROM,
+	CR_FD_FILES,
 	CR_FD_SK_QUEUES,
-	CR_FD_REG_FILES,
-	CR_FD_EXT_FILES,
-	CR_FD_NS_FILES,
-	CR_FD_INETSK,
-	CR_FD_UNIXSK,
-	CR_FD_PACKETSK,
-	CR_FD_NETLINK_SK,
-	CR_FD_PIPES,
 	CR_FD_PIPES_DATA,
-	CR_FD_FIFO,
 	CR_FD_FIFO_DATA,
-	CR_FD_TTY_FILES,
 	CR_FD_TTY_INFO,
 	CR_FD_TTY_DATA,
 	CR_FD_REMAP_FPATH,
-	CR_FD_EVENTFD_FILE,
-	CR_FD_EVENTPOLL_FILE,
-	CR_FD_SIGNALFD,
-	CR_FD_INOTIFY_FILE,
-	CR_FD_FANOTIFY_FILE,
-	CR_FD_TUNFILE,
 	CR_FD_CGROUP,
-	CR_FD_TIMERFD,
 	CR_FD_FILE_LOCKS,
 	CR_FD_SECCOMP,
 	_CR_FD_GLOB_TO,
@@ -86,8 +69,10 @@ enum {
 	CR_FD_TMPFS_IMG,
 	CR_FD_TMPFS_DEV,
 	CR_FD_BINFMT_MISC,
+	CR_FD_BINFMT_MISC_OLD,
 	CR_FD_PAGES,
 
+	CR_FD_SIGACT,
 	CR_FD_VMAS,
 	CR_FD_PAGES_OLD,
 	CR_FD_SHM_PAGES_OLD,
@@ -104,6 +89,23 @@ enum {
 	CR_FD_INOTIFY_WD,
 	CR_FD_FANOTIFY_MARK,
 	CR_FD_EVENTPOLL_TFD,
+	CR_FD_REG_FILES,
+	CR_FD_INETSK,
+	CR_FD_NS_FILES,
+	CR_FD_PACKETSK,
+	CR_FD_NETLINK_SK,
+	CR_FD_EVENTFD_FILE,
+	CR_FD_EVENTPOLL_FILE,
+	CR_FD_SIGNALFD,
+	CR_FD_TUNFILE,
+	CR_FD_TIMERFD,
+	CR_FD_INOTIFY_FILE,
+	CR_FD_FANOTIFY_FILE,
+	CR_FD_EXT_FILES,
+	CR_FD_UNIXSK,
+	CR_FD_FIFO,
+	CR_FD_PIPES,
+	CR_FD_TTY_FILES,
 
 	CR_FD_AUTOFS,
 
