@@ -1,7 +1,13 @@
 #ifndef __CR_KERNDAT_H__
 #define __CR_KERNDAT_H__
 
-#include "asm/types.h"
+#include <stdbool.h>
+
+#include "int.h"
+#include "config.h"
+#ifdef CONFIG_VDSO
+#include "util-vdso.h"
+#endif
 
 struct stat;
 
@@ -11,10 +17,9 @@ struct stat;
  */
 
 extern int kerndat_init(void);
-extern int kerndat_init_rst(void);
 extern int kerndat_get_dirty_track(void);
 extern int kerndat_fdinfo_has_lock(void);
-extern int kerndat_loginuid(bool only_dump);
+extern int kerndat_loginuid(void);
 
 enum pagemap_func {
 	PM_UNKNOWN,
@@ -23,9 +28,15 @@ enum pagemap_func {
 	PM_FULL,
 };
 
+enum loginuid_func {
+	LUID_NONE,
+	LUID_READ,
+	LUID_FULL,
+};
+
 struct kerndat_s {
+	u32 magic1, magic2;
 	dev_t shmem_dev;
-	int tcp_max_rshare;
 	int last_cap;
 	u64 zero_page_pfn;
 	bool has_dirty_track;
@@ -33,8 +44,25 @@ struct kerndat_s {
 	bool has_fdinfo_lock;
 	unsigned long task_size;
 	bool ipv6;
-	bool has_loginuid;
+	enum loginuid_func luid;
+	bool compat_cr;
 	enum pagemap_func pmap;
+	unsigned int has_xtlocks;
+	unsigned long mmap_min_addr;
+	bool has_tcp_half_closed;
+	bool stack_guard_gap_hidden;
+	int lsm;
+	bool has_uffd;
+	unsigned long uffd_features;
+	bool has_thp_disable;
+	bool can_map_vdso;
+	bool vdso_hint_reliable;
+#ifdef CONFIG_VDSO
+	struct vdso_symtable	vdso_sym;
+#ifdef CONFIG_COMPAT
+	struct vdso_symtable	vdso_sym_compat;
+#endif
+#endif
 };
 
 extern struct kerndat_s kdat;
@@ -55,4 +83,7 @@ enum {
  */
 extern int kerndat_fs_virtualized(unsigned int which, u32 kdev);
 
+extern int kerndat_tcp_repair();
+extern int kerndat_uffd(void);
+
 #endif /* __CR_KERNDAT_H__ */
