@@ -3,32 +3,13 @@
 
 #include <sys/types.h>
 
-#include "asm/types.h"
-#include "list.h"
+#include "common/list.h"
 
 struct proc_mountinfo;
 struct pstree_item;
 struct fstype;
 struct ns_id;
 
-/*
- * Structure to keep external mount points resolving info.
- *
- * On dump the key is the mountpoint as seen from the mount
- * namespace, the val is some name that will be put into image
- * instead of the mount point's root path.
- *
- * On restore the key is the name from the image (the one
- * mentioned above) and the val is the path in criu's mount
- * namespace that will become the mount point's root, i.e. --
- * be bind mounted to the respective mountpoint.
- */
-struct ext_mount {
-	struct list_head	list;
-	char			*key;
-	char			*val;
-};
-
 #define MOUNT_INVALID_DEV	(0)
 
 struct mount_info {
@@ -68,7 +49,7 @@ struct mount_info {
 	struct mount_info	*next;
 	struct ns_id		*nsid;
 
-	struct ext_mount	*external;
+	char			*external;
 	bool			internal_sharing;
 
 	/* tree linkage */
@@ -90,6 +71,11 @@ struct mount_info {
 
 extern struct mount_info *mntinfo;
 extern struct ns_desc mnt_ns_desc;
+#ifdef CONFIG_BINFMT_MISC_VIRTUALIZED
+extern int collect_binfmt_misc(void);
+#else
+static inline int collect_binfmt_misc(void) { return 0; }
+#endif
 
 extern struct mount_info *mnt_entry_alloc();
 extern void mnt_entry_free(struct mount_info *mi);
@@ -101,8 +87,8 @@ extern struct ns_id *lookup_nsid_by_mnt_id(int mnt_id);
 
 extern int open_mount(unsigned int s_dev);
 extern int __open_mountpoint(struct mount_info *pm, int mnt_fd);
-extern struct fstype *find_fstype_by_name(char *fst);
-extern bool add_fsname_auto(const char *names);
+extern int mnt_is_dir(struct mount_info *pm);
+extern int open_mountpoint(struct mount_info *pm);
 
 extern struct mount_info *collect_mntinfo(struct ns_id *ns, bool for_dump);
 extern int prepare_mnt_ns(void);
@@ -124,33 +110,16 @@ extern int depopulate_roots_yard(int mntns_root, bool clean_remaps);
 
 extern int rst_get_mnt_root(int mnt_id, char *path, int plen);
 extern int ext_mount_add(char *key, char *val);
+extern int ext_mount_parse_auto(char *key);
 extern int mntns_maybe_create_roots(void);
 extern int read_mnt_ns_img(void);
 extern void cleanup_mnt_ns(void);
-
-struct mount_info;
-typedef int (*mount_fn_t)(struct mount_info *mi, const char *src, const
-			  char *fstype, unsigned long mountflags);
-
-struct fstype {
-	char *name;
-	int code;
-	int (*dump)(struct mount_info *pm);
-	int (*restore)(struct mount_info *pm);
-	int (*parse)(struct mount_info *pm);
-	mount_fn_t mount;
-};
+extern void clean_cr_time_mounts(void);
 
 extern bool add_skip_mount(const char *mountpoint);
 struct ns_id;
 extern struct mount_info *parse_mountinfo(pid_t pid, struct ns_id *nsid, bool for_dump);
 
-/* callback for AUFS support */
-extern int aufs_parse(struct mount_info *mi);
-
-/* callback for OverlayFS support */
-extern int overlayfs_parse(struct mount_info *mi);
-
 extern int check_mnt_id(void);
 
 #endif /* __CR_MOUNT_H__ */
