@@ -2,8 +2,8 @@
 #define __CR_PROC_PARSE_H__
 
 #include <sys/types.h>
-#include "asm/types.h"
 
+#include <compel/compel.h>
 #include "images/seccomp.pb-c.h"
 
 #define PROC_TASK_COMM_LEN	32
@@ -73,15 +73,11 @@ struct seccomp_info {
 #define PROC_CAP_SIZE	2
 
 struct proc_status_creds {
+	struct seize_task_status s;
+
 	unsigned int uids[4];
 	unsigned int gids[4];
 
-	char			state;
-	int			ppid;
-	unsigned long long	sigpnd;
-	unsigned long long	shdpnd;
-
-	int			seccomp_mode;
 	u32			last_filter;
 
 	/*
@@ -94,16 +90,13 @@ struct proc_status_creds {
 	u32			cap_bnd[PROC_CAP_SIZE];
 };
 
-bool proc_status_creds_dumpable(struct proc_status_creds *parent,
-				struct proc_status_creds *child);
-
 #define INVALID_UID ((uid_t)-1)
 
 extern int parse_pid_stat(pid_t pid, struct proc_pid_stat *s);
 extern unsigned int parse_pid_loginuid(pid_t pid, int *err, bool ignore_noent);
 extern int parse_pid_oom_score_adj(pid_t pid, int *err);
 extern int prepare_loginuid(unsigned int value, unsigned int loglevel);
-extern int parse_pid_status(pid_t pid, struct proc_status_creds *);
+extern int parse_pid_status(pid_t pid, struct seize_task_status *, void *data);
 extern int parse_file_locks(void);
 extern int get_fd_mntid(int fd, int *mnt_id);
 
@@ -112,4 +105,7 @@ extern int parse_threads(int pid, struct pid **_t, int *_n);
 
 int parse_children(pid_t pid, pid_t **_c, int *_n);
 
+extern bool is_vma_range_fmt(char *line);
+extern void parse_vmflags(char *buf, u32 *flags, u64 *madv, int *io_pf);
+
 #endif /* __CR_PROC_PARSE_H__ */
