@@ -20,6 +20,9 @@ enum sfd_type {
 	USERNSD_SK,	/* Socket for usernsd */
 	NS_FD_OFF,	/* Node's net namespace fd */
 	TRANSPORT_FD_OFF, /* to transfer file descriptors */
+	RPC_SK_OFF,
+	FDSTORE_SK_OFF,
+	LAZY_PAGES_SK_OFF, /* socket for communication with lazy-pages daemon */
 
 	SERVICE_FD_MAX
 };
