@@ -1,3 +1,4 @@
+#include <stdio.h>
 #include <unistd.h>
 #include <sys/socket.h>
 #include <linux/if.h>
@@ -17,6 +18,7 @@
 #include "tun.h"
 #include "net.h"
 #include "namespaces.h"
+#include "xmalloc.h"
 
 #include "images/tun.pb-c.h"
 
@@ -266,6 +268,7 @@ static int dump_tunfile(int lfd, u32 id, const struct fd_parms *p)
 {
 	int ret;
 	struct cr_img *img;
+	FileEntry fe = FILE_ENTRY__INIT;
 	TunfileEntry tfe = TUNFILE_ENTRY__INIT;
 	struct ifreq ifr;
 
@@ -304,8 +307,12 @@ static int dump_tunfile(int lfd, u32 id, const struct fd_parms *p)
 			return -1;
 	}
 
-	img = img_from_set(glob_imgset, CR_FD_TUNFILE);
-	return pb_write_one(img, &tfe, PB_TUNFILE);
+	fe.type = FD_TYPES__TUNF;
+	fe.id = tfe.id;
+	fe.tunf = &tfe;
+
+	img = img_from_set(glob_imgset, CR_FD_FILES);
+	return pb_write_one(img, &fe, PB_FILE);
 }
 
 const struct fdtype_ops tunfile_dump_ops = {
@@ -318,7 +325,7 @@ struct tunfile_info {
 	TunfileEntry *tfe;
 };
 
-static int tunfile_open(struct file_desc *d)
+static int tunfile_open(struct file_desc *d, int *new_fd)
 {
 	int fd;
 	struct tunfile_info *ti;
@@ -332,7 +339,7 @@ static int tunfile_open(struct file_desc *d)
 
 	if (!ti->tfe->netdev)
 		/* just-opened tun file */
-		return fd;
+		goto ok;;
 
 	tl = find_tun_link(ti->tfe->netdev);
 	if (!tl) {
@@ -365,8 +372,9 @@ static int tunfile_open(struct file_desc *d)
 			goto err;
 		}
 	}
-
-	return fd;
+ok:
+	*new_fd = fd;
+	return 0;
 
 err:
 	close(fd);
@@ -397,7 +405,7 @@ struct collect_image_info tunfile_cinfo = {
 	.collect = collect_one_tunfile,
 };
 
-int dump_tun_link(NetDeviceEntry *nde, struct cr_imgset *fds)
+int dump_tun_link(NetDeviceEntry *nde, struct cr_imgset *fds, struct nlattr **info)
 {
 	TunLinkEntry tle = TUN_LINK_ENTRY__INIT;
 	char spath[64];
@@ -428,7 +436,7 @@ int dump_tun_link(NetDeviceEntry *nde, struct cr_imgset *fds)
 	tle.sndbuf = tl->dmp.sndbuf;
 
 	nde->tun = &tle;
-	return write_netdev_img(nde, fds);
+	return write_netdev_img(nde, fds, info);
 }
 
 int restore_one_tun(NetDeviceEntry *nde, int nlsk)
