@@ -8,7 +8,7 @@ chmod 0777 test
 chmod 0777 test/zdtm/transition/
 chmod 0777 test/zdtm/static
 
-./test/zdtm.py run -a --keep-going -k always --parallel 4 -x zdtm/static/rtc
+./test/zdtm.py run -a --keep-going -k always --parallel 4 -x zdtm/static/rtc "$@"
 ret=$?
 
 for i in `find / -name 'asan.log*'`; do
