@@ -2,5 +2,5 @@
 
 source `dirname $0`/criu-lib.sh &&
 prep &&
-make -C test -j 4 ZDTM_ARGS="-C -x '\(maps04\|mountpoints\|inotify_irmap\)'" &&
+make -C test -j 4 ZDTM_ARGS="-C -x '\(maps04\|mountpoints\|inotify_irmap\)'" zdtm &&
 true || fail
