@@ -1,5 +1,5 @@
 #!/usr/bin/env python2
-# vim: noet
+# vim: noet ts=8 sw=8 sts=8
 import argparse
 import glob
 import os
@@ -61,6 +61,7 @@ def make_tests_root():
 		atexit.register(clean_tests_root)
 	return tests_root[1]
 
+
 # Report generation
 
 report_dir = None
@@ -82,8 +83,9 @@ def add_to_report(path, tgt_name):
 			tgt_path = os.path.join(report_dir, tgt_name + ".%d" % att)
 			att += 1
 
+		ignore = shutil.ignore_patterns('*.socket')
 		if os.path.isdir(path):
-			shutil.copytree(path, tgt_path)
+			shutil.copytree(path, tgt_path, ignore = ignore)
 		else:
 			if not os.path.exists(os.path.dirname(tgt_path)):
 				os.mkdir(os.path.dirname(tgt_path))
@@ -123,6 +125,7 @@ def check_core_files():
 
 	return True
 
+
 # Arch we run on
 arch = os.uname()[4]
 
@@ -246,8 +249,7 @@ class ns_flavor:
 
 	def fini(self):
 		if self.root_mounted:
-			subprocess.check_call(["mount", "--make-private", self.root])
-			subprocess.check_call(["umount", "-l", self.root])
+			subprocess.check_call(["./umount2", self.root])
 			self.root_mounted = False
 
 	@staticmethod
@@ -290,10 +292,10 @@ def encode_flav(f):
 
 
 def decode_flav(i):
-	i = i - 128
-	if i in flavors:
+	try:
 		return flavors.keys()[i - 128]
-	return "unknown"
+	except:
+		return "unknown"
 
 
 def tail(path):
@@ -362,6 +364,7 @@ class zdtm_test:
 		self._env = {}
 		self._deps = desc.get('deps', [])
 		self.auto_reap = True
+		self.__timeout = int(self.__desc.get('timeout') or 30)
 
 	def __make_action(self, act, env = None, root = None):
 		sys.stdout.flush()  # Not to let make's messages appear before ours
@@ -375,6 +378,8 @@ class zdtm_test:
 
 		s = subprocess.Popen(s_args, env = env, cwd = root, close_fds = True,
 				preexec_fn = self.__freezer and self.__freezer.attach or None)
+		if act == "pid":
+			try_run_hook(self, ["--post-start"])
 		s.wait()
 
 		if self.__freezer:
@@ -384,7 +389,7 @@ class zdtm_test:
 		return self.__name + '.pid'
 
 	def __wait_task_die(self):
-		wait_pid_die(int(self.__pid), self.__name)
+		wait_pid_die(int(self.__pid), self.__name, self.__timeout)
 
 	def __add_wperms(self):
 		# Add write perms for .out and .pid files
@@ -510,6 +515,8 @@ class zdtm_test:
 
 	@staticmethod
 	def available():
+		if not os.access("umount2", os.X_OK):
+			subprocess.check_call(["make", "umount2"])
 		if not os.access("zdtm_ct", os.X_OK):
 			subprocess.check_call(["make", "zdtm_ct"])
 		if not os.access("zdtm/lib/libzdtmtst.a", os.F_OK):
@@ -670,12 +677,16 @@ join_ns_file = '/run/netns/zdtm_netns'
 
 class criu_cli:
 	@staticmethod
-	def run(action, args, fault = None, strace = [], preexec = None):
-		env = None
+	def run(action, args, fault = None, strace = [], preexec = None, nowait = False):
+		env = dict(os.environ, ASAN_OPTIONS = "log_path=asan.log:disable_coredump=0:detect_leaks=0")
+
 		if fault:
 			print "Forcing %s fault" % fault
-			env = dict(os.environ, CRIU_FAULT = fault)
+			env['CRIU_FAULT'] = fault
+
 		cr = subprocess.Popen(strace + [criu_bin, action] + args, env = env, preexec_fn = preexec)
+		if nowait:
+			return cr
 		return cr.wait()
 
 
@@ -710,40 +721,51 @@ class criu_rpc:
 			if arg == '--root':
 				criu.opts.root = args.pop(0)
 				continue
+			if arg == '--external':
+				criu.opts.external.append(args.pop(0))
+				continue
 
 			raise test_fail_exc('RPC for %s required' % arg)
 
 	@staticmethod
-	def run(action, args, fault = None, strace = [], preexec = None):
+	def run(action, args, fault = None, strace = [], preexec = None, nowait = False):
 		if fault:
 			raise test_fail_exc('RPC and FAULT not supported')
 		if strace:
 			raise test_fail_exc('RPC and SAT not supported')
 		if preexec:
 			raise test_fail_exc('RPC and PREEXEC not supported')
+		if nowait:
+			raise test_fail_exc("RPC and status-fd not supported")
 
 		ctx = {}  # Object used to keep info untill action is done
 		criu = crpc.criu()
 		criu.use_binary(criu_bin)
 		criu_rpc.__set_opts(criu, args, ctx)
 
-		if action == 'dump':
-			criu.dump()
-		elif action == 'restore':
-			if 'rd' not in ctx:
-				raise test_fail_exc('RPC Non-detached restore is impossible')
+		try:
+			if action == 'dump':
+				criu.dump()
+			elif action == 'restore':
+				if 'rd' not in ctx:
+					raise test_fail_exc('RPC Non-detached restore is impossible')
 
-			res = criu.restore()
-			pidf = ctx.get('pidf')
-			if pidf:
-				open(pidf, 'w').write('%d\n' % res.pid)
+				res = criu.restore()
+				pidf = ctx.get('pidf')
+				if pidf:
+					open(pidf, 'w').write('%d\n' % res.pid)
+			else:
+				raise test_fail_exc('RPC for %s required' % action)
+		except crpc.CRIUExceptionExternal:
+			print "Fail"
+			ret = -1
 		else:
-			raise test_fail_exc('RPC for %s required' % action)
+			ret = 0
 
 		imgd = ctx.get('imgd')
 		if imgd:
 			os.close(imgd)
-		return 0
+		return ret
 
 
 class criu:
@@ -753,15 +775,35 @@ class criu:
 		self.__iter = 0
 		self.__prev_dump_iter = None
 		self.__page_server = (opts['page_server'] and True or False)
+		self.__remote_lazy_pages = (opts['remote_lazy_pages'] and True or False)
+		self.__lazy_pages = (self.__remote_lazy_pages or
+				     opts['lazy_pages'] and True or False)
 		self.__restore_sibling = (opts['sibling'] and True or False)
 		self.__join_ns = (opts['join_ns'] and True or False)
+		self.__empty_ns = (opts['empty_ns'] and True or False)
 		self.__fault = (opts['fault'])
 		self.__script = opts['script']
 		self.__sat = (opts['sat'] and True or False)
 		self.__dedup = (opts['dedup'] and True or False)
+		self.__mdedup = (opts['noauto_dedup'] and True or False)
 		self.__user = (opts['user'] and True or False)
 		self.__leave_stopped = (opts['stop'] and True or False)
 		self.__criu = (opts['rpc'] and criu_rpc or criu_cli)
+		self.__lazy_pages_p = None
+		self.__page_server_p = None
+
+	def fini(self):
+		if self.__lazy_pages_p:
+			ret = self.__lazy_pages_p.wait()
+			self.__lazy_pages_p = None
+			if ret:
+				raise test_fail_exc("criu lazy-pages exited with %s" % ret)
+		if self.__page_server_p:
+			ret = self.__page_server_p.wait()
+			self.__page_server_p = None
+			if ret:
+				raise test_fail_exc("criu page-server exited with %s" % ret)
+                return
 
 	def logs(self):
 		return self.__dump_path
@@ -793,7 +835,7 @@ class criu:
 		os.setresgid(58467, 58467, 58467)
 		os.setresuid(18943, 18943, 18943)
 
-	def __criu_act(self, action, opts, log = None):
+	def __criu_act(self, action, opts = [], log = None, nowait = False):
 		if not log:
 			log = action + ".log"
 
@@ -822,7 +864,23 @@ class criu:
 
 		__ddir = self.__ddir()
 
-		ret = self.__criu.run(action, s_args, self.__fault, strace, preexec)
+		status_fds = None
+		if nowait:
+			status_fds = os.pipe()
+			s_args += ["--status-fd", str(status_fds[1])]
+
+		ns_last_pid = open("/proc/sys/kernel/ns_last_pid").read()
+
+		ret = self.__criu.run(action, s_args, self.__fault, strace, preexec, nowait)
+
+		if nowait:
+			os.close(status_fds[1])
+			if os.read(status_fds[0], 1) != '\0':
+				ret = ret.wait()
+				raise test_fail_exc("criu %s exited with %s" % (action, ret))
+			os.close(status_fds[0])
+			return ret
+
 		grep_errors(os.path.join(__ddir, log))
 		if ret != 0:
 			if self.__fault and int(self.__fault) < 128:
@@ -835,6 +893,9 @@ class criu:
 				else:
 					# on restore we move only a log file, because we need images
 					os.rename(os.path.join(__ddir, log), os.path.join(__ddir, log + ".fail"))
+				# restore ns_last_pid to avoid a case when criu gets
+				# PID of one of restored processes.
+				open("/proc/sys/kernel/ns_last_pid", "w+").write(ns_last_pid)
 				# try again without faults
 				print "Run criu " + action
 				ret = self.__criu.run(action, s_args, False, strace, preexec)
@@ -859,11 +920,11 @@ class criu:
 		if self.__page_server:
 			print "Adding page server"
 
-			ps_opts = ["--port", "12345", "--daemon", "--pidfile", "ps.pid"]
+			ps_opts = ["--port", "12345"]
 			if self.__dedup:
 				ps_opts += ["--auto-dedup"]
 
-			self.__criu_act("page-server", opts = ps_opts)
+			self.__page_server_p = self.__criu_act("page-server", opts = ps_opts, nowait = True)
 			a_opts += ["--page-server", "--address", "127.0.0.1", "--port", "12345"]
 
 		a_opts += self.__test.getdopts()
@@ -875,20 +936,27 @@ class criu:
 
 		criu_dir = os.path.dirname(os.getcwd())
 		if os.getenv("GCOV"):
-			a_opts.append("--ext-mount-map")
-			a_opts.append("%s:zdtm" % criu_dir)
+			a_opts.append('--external')
+			a_opts.append('mnt[%s]:zdtm' % criu_dir)
 
 		if self.__leave_stopped:
 			a_opts += ['--leave-stopped']
+		if self.__empty_ns:
+			a_opts += ['--empty-ns', 'net']
 
 		self.__criu_act(action, opts = a_opts + opts)
+		if self.__mdedup and self.__iter > 1:
+			self.__criu_act("dedup", opts = [])
 
 		if self.__leave_stopped:
 			pstree_check_stopped(self.__test.getpid())
 			pstree_signal(self.__test.getpid(), signal.SIGKILL)
 
-		if self.__page_server:
-			wait_pid_die(int(rpidfile(self.__ddir() + "/ps.pid")), "page server")
+		if self.__page_server_p:
+			ret = self.__page_server_p.wait()
+			self.__page_server_p = None
+			if ret:
+				raise test_fail_exc("criu page-server exited with %d" % ret)
 
 	def restore(self):
 		r_opts = []
@@ -899,12 +967,25 @@ class criu:
 		if self.__join_ns:
 			r_opts.append("--join-ns")
 			r_opts.append("net:%s" % join_ns_file)
+		if self.__empty_ns:
+			r_opts += ['--empty-ns', 'net']
+			r_opts += ['--action-script', os.getcwd() + '/empty-netns-prep.sh']
 
 		self.__prev_dump_iter = None
 		criu_dir = os.path.dirname(os.getcwd())
 		if os.getenv("GCOV"):
-			r_opts.append("--ext-mount-map")
-			r_opts.append("zdtm:%s" % criu_dir)
+			r_opts.append('--external')
+			r_opts.append('mnt[zdtm]:%s' % criu_dir)
+
+		if self.__lazy_pages:
+			lp_opts = []
+			if self.__remote_lazy_pages:
+				lp_opts += ['--page-server', "--port", "12345"]
+				ps_opts = ["--pidfile", "ps.pid",
+					   "--port", "12345", "--lazy-pages"]
+				self.__page_server_p = self.__criu_act("page-server", opts = ps_opts, nowait = True)
+			self.__lazy_pages_p = self.__criu_act("lazy-pages", opts = lp_opts, nowait = True)
+			r_opts += ["--lazy-pages"]
 
 		if self.__leave_stopped:
 			r_opts += ['--leave-stopped']
@@ -925,6 +1006,16 @@ class criu:
 			print "CRIU binary not built"
 			sys.exit(1)
 
+	def kill(self):
+		if self.__lazy_pages_p:
+			self.__lazy_pages_p.terminate()
+			print "criu lazy-pages exited with %s" & self.wait()
+			self.__lazy_pages_p = None
+		if self.__page_server_p:
+			self.__page_server_p.terminate()
+			print "criu page-server exited with %s" & self.wait()
+			self.__page_server_p = None
+
 
 def try_run_hook(test, args):
 	hname = test.getname() + '.hook'
@@ -934,6 +1025,7 @@ def try_run_hook(test, args):
 		if hook.wait() != 0:
 			raise test_fail_exc("hook " + " ".join(args))
 
+
 #
 # Step by step execution
 #
@@ -981,14 +1073,20 @@ def cr(cr_api, test, opts):
 
 		sbs('pre-dump')
 
+		os.environ["ZDTM_TEST_PID"] = str(test.getpid())
 		if opts['norst']:
+			try_run_hook(test, ["--pre-dump"])
 			cr_api.dump("dump", opts = ["--leave-running"])
 		else:
+			try_run_hook(test, ["--pre-dump"])
 			cr_api.dump("dump")
 			test.gone()
 			sbs('pre-restore')
 			try_run_hook(test, ["--pre-restore"])
 			cr_api.restore()
+			os.environ["ZDTM_TEST_PID"] = str(test.getpid())
+                        os.environ["ZDTM_IMG_DIR"] = cr_api.logs()
+			try_run_hook(test, ["--post-restore"])
 			sbs('post-restore')
 
 		time.sleep(iters[1])
@@ -1055,6 +1153,11 @@ def check_visible_state(test, state, opts):
 
 		old_maps = state[1][pid]
 		new_maps = new[1][pid]
+		if os.getenv("COMPAT_TEST"):
+			# the vsyscall vma isn't unmapped from x32 processes
+			vsyscall = "ffffffffff600000-ffffffffff601000 ['r-xp']"
+			if vsyscall in new_maps and vsyscall not in old_maps:
+				new_maps.remove(vsyscall)
 		if old_maps != new_maps:
 			print "%s: Old maps lost: %s" % (pid, old_maps - new_maps)
 			print "%s: New maps appeared: %s" % (pid, new_maps - old_maps)
@@ -1253,11 +1356,14 @@ def do_run_test(tname, tdesc, flavs, opts):
 				if opts['join_ns']:
 					check_joinns_state(t)
 				t.stop()
+				cr_api.fini()
 				try_run_hook(t, ["--clean"])
 		except test_fail_exc as e:
 			print_sep("Test %s FAIL at %s" % (tname, e.step), '#')
 			t.print_output()
 			t.kill()
+			cr_api.kill()
+			try_run_hook(t, ["--clean"])
 			if cr_api.logs():
 				add_to_report(cr_api.logs(), tname.replace('/', '_') + "_" + f + "/images")
 			if opts['keep_img'] == 'never':
@@ -1282,6 +1388,10 @@ class launcher:
 		self.__subs = {}
 		self.__fail = False
 		self.__file_report = None
+		self.__junit_file = None
+		self.__junit_test_cases = None
+		self.__failed = []
+		self.__nr_skip = 0
 		if self.__max > 1 and self.__total > 1:
 			self.__use_log = True
 		elif opts['report']:
@@ -1290,28 +1400,46 @@ class launcher:
 			self.__use_log = False
 
 		if opts['report'] and (opts['keep_going'] or self.__total == 1):
+			global TestSuite, TestCase
+			from junit_xml import TestSuite, TestCase
 			now = datetime.datetime.now()
 			att = 0
 			reportname = os.path.join(report_dir, "criu-testreport.tap")
-			while os.access(reportname, os.F_OK):
+			junitreport = os.path.join(report_dir, "criu-testreport.xml")
+			while os.access(reportname, os.F_OK) or os.access(junitreport, os.F_OK):
 				reportname = os.path.join(report_dir, "criu-testreport" + ".%d.tap" % att)
+				junitreport = os.path.join(report_dir, "criu-testreport" + ".%d.xml" % att)
 				att += 1
 
+			self.__junit_file = open(junitreport, 'a')
+			self.__junit_test_cases = []
+
 			self.__file_report = open(reportname, 'a')
+			print >> self.__file_report, "TAP version 13"
 			print >> self.__file_report, "# Hardware architecture: " + arch
 			print >> self.__file_report, "# Timestamp: " + now.strftime("%Y-%m-%d %H:%M") + " (GMT+1)"
 			print >> self.__file_report, "# "
-			print >> self.__file_report, "TAP version 13"
 			print >> self.__file_report, "1.." + str(nr_tests)
+		self.__taint = open("/proc/sys/kernel/tainted").read()
+		if int(self.__taint, 0) != 0:
+			print "The kernel is tainted: %r" % self.__taint
+			if not opts["ignore_taint"]:
+				raise Exception("The kernel is tainted: %r" % self.__taint)
 
-	def __show_progress(self):
+	def __show_progress(self, msg):
 		perc = self.__nr * 16 / self.__total
-		print "=== Run %d/%d %s" % (self.__nr, self.__total, '=' * perc + '-' * (16 - perc))
+		print "=== Run %d/%d %s %s" % (self.__nr, self.__total, '=' * perc + '-' * (16 - perc), msg)
 
 	def skip(self, name, reason):
 		print "Skipping %s (%s)" % (name, reason)
 		self.__nr += 1
 		self.__runtest += 1
+		self.__nr_skip += 1
+
+		if self.__junit_test_cases is not None:
+			tc = TestCase(name)
+			tc.add_skipped_info(reason)
+			self.__junit_test_cases.append(tc)
 		if self.__file_report:
 			testline = "ok %d - %s # SKIP %s" % (self.__runtest, name, reason)
 			print >> self.__file_report, testline
@@ -1321,15 +1449,19 @@ class launcher:
 		if len(self.__subs) >= self.__max:
 			self.wait()
 
+		taint = open("/proc/sys/kernel/tainted").read()
+		if self.__taint != taint:
+			raise Exception("The kernel is tainted: %r (%r)" % (taint, self.__taint))
+
 		if test_flag(desc, 'excl'):
 			self.wait_all()
 
 		self.__nr += 1
-		self.__show_progress()
+		self.__show_progress(name)
 
-		nd = ('nocr', 'norst', 'pre', 'iters', 'page_server', 'sibling', 'stop',
-				'fault', 'keep_img', 'report', 'snaps', 'sat', 'script', 'rpc',
-				'join_ns', 'dedup', 'sbs', 'freezecg', 'user', 'dry_run')
+		nd = ('nocr', 'norst', 'pre', 'iters', 'page_server', 'sibling', 'stop', 'empty_ns',
+				'fault', 'keep_img', 'report', 'snaps', 'sat', 'script', 'rpc', 'lazy_pages',
+				'join_ns', 'dedup', 'sbs', 'freezecg', 'user', 'dry_run', 'noauto_dedup', 'remote_lazy_pages')
 		arg = repr((name, desc, flavor, {d: self.__opts[d] for d in nd}))
 
 		if self.__use_log:
@@ -1342,7 +1474,7 @@ class launcher:
 		sub = subprocess.Popen(["./zdtm_ct", "zdtm.py"],
 				env = dict(os.environ, CR_CT_TEST_INFO = arg),
 				stdout = log, stderr = subprocess.STDOUT, close_fds = True)
-		self.__subs[sub.pid] = {'sub': sub, 'log': logf, 'name': name}
+		self.__subs[sub.pid] = {'sub': sub, 'log': logf, 'name': name, "start": time.time()}
 
 		if test_flag(desc, 'excl'):
 			self.wait()
@@ -1352,12 +1484,19 @@ class launcher:
 		self.__runtest += 1
 		if pid != 0:
 			sub = self.__subs.pop(pid)
+			tc = None
+			if self.__junit_test_cases is not None:
+				tc = TestCase(sub['name'], elapsed_sec=time.time() - sub['start'])
+				self.__junit_test_cases.append(tc)
 			if status != 0:
 				self.__fail = True
 				failed_flavor = decode_flav(os.WEXITSTATUS(status))
+				self.__failed.append([sub['name'], failed_flavor])
 				if self.__file_report:
 					testline = "not ok %d - %s # flavor %s" % (self.__runtest, sub['name'], failed_flavor)
-					details = {'output': open(sub['log']).read()}
+					output = open(sub['log']).read()
+					details = {'output': output}
+					tc.add_error_info(output = output)
 					print >> self.__file_report, testline
 					print >> self.__file_report, yaml.dump(details, explicit_start=True, explicit_end=True, default_style='|')
 				if sub['log']:
@@ -1397,7 +1536,20 @@ class launcher:
 		if not opts['fault'] and check_core_files():
 			self.__fail = True
 		if self.__file_report:
+			ts = TestSuite(opts['title'], self.__junit_test_cases, os.getenv("NODE_NAME"))
+			self.__junit_file.write(TestSuite.to_xml_string([ts]))
 			self.__file_report.close()
+
+		if opts['keep_going']:
+			if self.__fail:
+				print_sep("%d TEST(S) FAILED (TOTAL %d/SKIPPED %d)"
+						% (len(self.__failed), self.__total, self.__nr_skip), "#")
+				for failed in self.__failed:
+					print " * %s(%s)" % (failed[0], failed[1])
+			else:
+				print_sep("ALL TEST(S) PASSED (TOTAL %d/SKIPPED %d)"
+						% (self.__total, self.__nr_skip), "#")
+
 		if self.__fail:
 			print_sep("FAIL", "#")
 			sys.exit(1)
@@ -1405,16 +1557,25 @@ class launcher:
 
 def all_tests(opts):
 	desc = eval(open(opts['set'] + '.desc').read())
-	lst = subprocess.Popen(['find', desc['dir'], '-type', 'f', '-executable'],
-			stdout = subprocess.PIPE)
+
+	files = []
+	mask = stat.S_IFREG | stat.S_IXUSR
+	for d in os.walk(desc['dir']):
+		for f in d[2]:
+			fp = os.path.join(d[0], f)
+			st = os.lstat(fp)
+			if (st.st_mode & mask) != mask:
+				continue
+			if stat.S_IFMT(st.st_mode) in [stat.S_IFLNK, stat.S_IFSOCK]:
+				continue
+			files.append(fp)
 	excl = map(lambda x: os.path.join(desc['dir'], x), desc['exclude'])
 	tlist = filter(lambda x:
 			not x.endswith('.checkskip') and
 			not x.endswith('.hook') and
 			x not in excl,
-			map(lambda x: x.strip(), lst.stdout.readlines())
+			map(lambda x: x.strip(), files)
 			)
-	lst.wait()
 	return tlist
 
 
@@ -1424,7 +1585,7 @@ default_test = {}
 
 def get_test_desc(tname):
 	d_path = tname + '.desc'
-	if os.access(d_path, os.F_OK):
+	if os.access(d_path, os.F_OK) and os.path.getsize(d_path) > 0:
 		return eval(open(d_path).read())
 
 	return default_test
@@ -1447,15 +1608,33 @@ def print_sep(title, sep = "=", width = 80):
 	print (" " + title + " ").center(width, sep)
 
 
+def print_error(line):
+	line = line.rstrip()
+	print line
+	if line.endswith('>'):  # combine pie output
+		return True
+	return False
+
+
 def grep_errors(fname):
 	first = True
+	print_next = False
+	before = []
 	for l in open(fname):
+		before.append(l)
+		if len(before) > 5:
+			before.pop(0)
 		if "Error" in l:
 			if first:
 				print_fname(fname, 'log')
 				print_sep("grep Error", "-", 60)
 				first = False
-			print l,
+			for i in before:
+				print_next = print_error(i)
+			before = []
+		else:
+			if print_next:
+				print_next = print_error(l)
 	if not first:
 		print_sep("ERROR OVER", "-", 60)
 
@@ -1469,20 +1648,15 @@ def run_tests(opts):
 			print "Tracking memory is not available"
 			return
 
-	if opts['keep_going'] and (not opts['all']):
-		print "[WARNING] Option --keep-going is more useful with option --all."
-
 	if opts['all']:
 		torun = all_tests(opts)
 		run_all = True
 	elif opts['tests']:
 		r = re.compile(opts['tests'])
 		torun = filter(lambda x: r.match(x), all_tests(opts))
-		opts['keep_going'] = False
 		run_all = True
 	elif opts['test']:
 		torun = opts['test']
-		opts['keep_going'] = False
 		run_all = False
 	elif opts['from']:
 		if not os.access(opts['from'], os.R_OK):
@@ -1496,6 +1670,10 @@ def run_tests(opts):
 		print "Specify test with -t <name> or -a"
 		return
 
+	if opts['keep_going'] and len(torun) < 2:
+		print "[WARNING] Option --keep-going is more useful when running multiple tests"
+		opts['keep_going'] = False
+
 	if opts['exclude']:
 		excl = re.compile(".*(" + "|".join(opts['exclude']) + ")")
 		print "Compiled exclusion list"
@@ -1513,6 +1691,15 @@ def run_tests(opts):
 		if subprocess.Popen(["ip", "netns", "exec", "zdtm_netns", "ip", "link", "set", "up", "dev", "lo"]).wait():
 			raise Exception("ip link set up dev lo")
 
+	if opts['lazy_pages'] or opts['remote_lazy_pages']:
+		uffd = criu.check("uffd")
+		uffd_noncoop = criu.check("uffd-noncoop")
+		if not uffd:
+			raise Exception("UFFD is not supported, cannot run with --lazy-pages")
+		if not uffd_noncoop:
+			# Most tests will work with 4.3 - 4.11
+			print "[WARNING] Non-cooperative UFFD is missing, some tests might spuriously fail"
+
 	l = launcher(opts, len(torun))
 	try:
 		for t in torun:
@@ -1527,6 +1714,10 @@ def run_tests(opts):
 				l.skip(t, "arch %s" % tdesc['arch'])
 				continue
 
+			if test_flag(tdesc, 'reqrst') and opts['norst']:
+				l.skip(t, "restore stage is required")
+				continue
+
 			if run_all and test_flag(tdesc, 'noauto'):
 				l.skip(t, "manual run only")
 				continue
@@ -1558,6 +1749,16 @@ def run_tests(opts):
 					l.skip(t, "samens test in the same namespace")
 					continue
 
+			if opts['lazy_pages'] or opts['remote_lazy_pages']:
+				if test_flag(tdesc, 'nolazy'):
+					l.skip(t, "lazy pages are not supported")
+					continue
+
+			if opts['remote_lazy_pages']:
+				if test_flag(tdesc, 'noremotelazy'):
+					l.skip(t, "remote lazy pages are not supported")
+					continue
+
 			test_flavs = tdesc.get('flavor', 'h ns uns').split()
 			opts_flavs = (opts['flavor'] or 'h,ns,uns').split(',')
 			if opts_flavs != ['best']:
@@ -1573,6 +1774,8 @@ def run_tests(opts):
 			# remove ns and uns flavor in join_ns
 			if opts['join_ns']:
 				run_flavs -= set(['ns', 'uns'])
+			if opts['empty_ns']:
+				run_flavs -= set(['h'])
 
 			if run_flavs:
 				l.run_test(t, tdesc, run_flavs)
@@ -1581,7 +1784,8 @@ def run_tests(opts):
 	finally:
 		l.finish()
 		if opts['join_ns']:
-			subprocess.Popen(["ip", "netns", "delete", "zdtm_netns"])
+			subprocess.Popen(["ip", "netns", "delete", "zdtm_netns"]).wait()
+
 
 sti_fmt = "%-40s%-10s%s"
 
@@ -1764,9 +1968,11 @@ rp.add_argument("-x", "--exclude", help = "Exclude tests from --all run", action
 
 rp.add_argument("--sibling", help = "Restore tests as siblings", action = 'store_true')
 rp.add_argument("--join-ns", help = "Restore tests and join existing namespace", action = 'store_true')
+rp.add_argument("--empty-ns", help = "Restore tests in empty net namespace", action = 'store_true')
 rp.add_argument("--pre", help = "Do some pre-dumps before dump (n[:pause])")
 rp.add_argument("--snaps", help = "Instead of pre-dumps do full dumps", action = 'store_true')
 rp.add_argument("--dedup", help = "Auto-deduplicate images on iterations", action = 'store_true')
+rp.add_argument("--noauto-dedup", help = "Manual deduplicate images on iterations", action = 'store_true')
 rp.add_argument("--nocr", help = "Do not CR anything, just check test works", action = 'store_true')
 rp.add_argument("--norst", help = "Don't restore tasks, leave them running after dump", action = 'store_true')
 rp.add_argument("--stop", help = "Check that --leave-stopped option stops ps tree.", action = 'store_true')
@@ -1786,6 +1992,10 @@ rp.add_argument("-k", "--keep-img", help = "Whether or not to keep images after
 		choices = ['always', 'never', 'failed'], default = 'failed')
 rp.add_argument("--report", help = "Generate summary report in directory")
 rp.add_argument("--keep-going", help = "Keep running tests in spite of failures", action = 'store_true')
+rp.add_argument("--ignore-taint", help = "Don't care about a non-zero kernel taint flag", action = 'store_true')
+rp.add_argument("--lazy-pages", help = "restore pages on demand", action = 'store_true')
+rp.add_argument("--remote-lazy-pages", help = "simulate lazy migration", action = 'store_true')
+rp.add_argument("--title", help = "A test suite title", default = "criu")
 
 lp = sp.add_parser("list", help = "List tests")
 lp.set_defaults(action = list_tests)
